/*
 * Constructs a newc-format cpio archive.
 *
 * Copyright 2012 Intel Corporation.
 *
 * This file is a derivative of usr/gen_init_cpio.c from within the
 * Linux kernel source distribution, version 2.6.34; it has been heavily
 * modified (starting in March 2012) for use within the Intel MIC
 * Platform Software Stack.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
 * USA.
 *
 * Authorship notices present in usr/gen_init_cpio.c:
 *
 * Original work by Jeff Garzik
 *
 * External file lists, symlink, pipe and fifo support by Thayne Harbaugh
 * Hard link support by Luciano Rocha
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <libgen.h>
#include <pthread.h>
#include <stdarg.h>
#include <time.h>
#include <errno.h>
#include <sys/dir.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/utsname.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>
#include <signal.h>
#include <limits.h>
#include <syslog.h>
#include <getopt.h>
#include <fnmatch.h>
#include <linux/kdev_t.h>
#include "mpssconfig.h"
#include "libmpsscommon.h"

#define SUPPORT_FILELIST

#define MAX_LINE ((2 * PATH_MAX) + 64)

struct filetree {
	int		type;
	char		*name;
	char		*source;
	char		*link;
	int		mode;
	int		uid;
	int		gid;
	char		inode_type;
	unsigned int	major;
	unsigned int	minor;
	struct filetree	*subdir;
	struct filetree *next;
	struct filetree *prev;
};

struct type_list {
	char	*name;
	int	len;
	int	type;
	void	(*func)(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs);
};

//#define MPSSDEBUG

#ifdef MPSSDEBUG
void dump_tree(struct filetree *elem, int level);
#endif

#define MPSS_UNKNOWN	0
#define MPSS_DIR	1
#define MPSS_FILE	2
#define MPSS_LINK	3
#define MPSS_DNODE	4
#define MPSS_PIPE	5
#define MPSS_SOCK	6

void add_dir(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs);
void add_file(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs);
void add_link(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs);
void add_dnode(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs);
void add_pipe(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs);
void add_sock(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs);

int gen_cpio(struct mpss_env *menv, struct mic_info *mic, struct filetree *top, struct mpss_elist *perrs);
int copy_nfs(struct mpss_env *menv, struct mic_info *mic, struct filetree *top, int createusr, struct mpss_elist *perrs);

void add_simple_dir(int id, char *source, char *target, char *lead, struct filetree *top, struct mpss_elist *perrs);
void add_simple_file(int id, char *source, char *target, char *lead, struct filetree *top, struct mpss_elist *perrs);
void add_simple_link(int id, char *source, char *target, char *lead, struct filetree *top, struct mpss_elist *perrs);
void add_simple_nod(int id, char type, char *source, char *target, char *lead, struct filetree *top, struct mpss_elist *perrs);
void add_simple_pipe(int id, char *source, char *target, char *lead, struct filetree *top, struct mpss_elist *perrs);
void add_simple_sock(int id, char *source, char *target, char *lead, struct filetree *top, struct mpss_elist *perrs);

void insert_list(int id, char *path, struct filetree *new, struct filetree *elem, char *source, struct mpss_elist *perrs);
void gunzip_image(char *name);

#define xstr(s) #s
#define str(s) xstr(s)

int
unzip_base_cpio(char *name, char *zfile, char *ofile, struct mpss_elist *perrs)
{
	struct stat sbuf;
	char buf[4096];
	int len;
	int rlen;
	int ifd;
	int ofd;

	if ((stat(zfile, &sbuf) != 0) ||
	    ((ifd = open(zfile, O_RDONLY)) < 0)) {
		add_perr(perrs, PERROR, "%s: error opening base cpio image '%s': %s", name, zfile, strerror(errno));
		return 1;
	}

	len = sbuf.st_size;

	if (stat(ofile, &sbuf) == 0)
		unlink(ofile);

	if ((ofd = open(ofile, O_WRONLY | O_CREAT, 0600)) < 0) {
		add_perr(perrs, PERROR, "%s: error opening image tmpfile '%s': %s", name, ofile);
		close(ifd);
		return 1;
	}

	while (len > 0) {
		rlen = read(ifd, buf, 4096);
		write(ofd, buf, rlen);
		len -= 4096;
	}

	close(ofd);
	close(ifd);

	gunzip_image(ofile);
	return 0;
}

struct type_list typelist[] = {
	{"dir", 3, MPSS_DIR, add_dir},
	{"file", 4, MPSS_FILE, add_file},
	{"slink", 5, MPSS_LINK, add_link},
	{"nod", 3, MPSS_DNODE, add_dnode},
	{"pipe", 4, MPSS_PIPE, add_pipe},
	{"sock", 4, MPSS_SOCK, add_sock},
	{NULL, 0, 0, NULL}
};

char *
get_type_string(int filetype)
{
	if ((filetype >= MPSS_DIR) && (filetype <= MPSS_SOCK))
		return typelist[filetype - 1].name;

	return "Unknown";
}

int
handle_dir(int id, struct filetree *top, char *sourcedir, char *listfile, struct mpss_elist *perrs)
{
	struct type_list *t;
	FILE *fp;
	char line[MAX_LINE];

	if ((fp = fopen(listfile, "r")) == NULL) {
		add_perr(perrs, PERROR, "mic%d failed to open '%s': %s", id, listfile, strerror(errno));
		return 1;
	}

	while (fgets(line, MAX_LINE, fp)) {
		// Simply skip any line that does not start with an interesting value
		t = &typelist[0];
		while (t->name != NULL) {
			if (!strncmp(line, t->name, t->len)) {
				t->func(id, &line[t->len + 1], top, sourcedir, perrs);
			}

			t++;
		}
	}

	fclose(fp);
	return 0;
}

int
handle_file(int id, struct filetree *top, char *source, char *dest, struct mpss_elist *perrs)
{
	struct stat sbuf;
	struct filetree *new;

	if (lstat(source, &sbuf) != 0)
		return 0;

	if (!S_ISREG(sbuf.st_mode))
		return 0;

	if ((new = malloc(sizeof(struct filetree))) == NULL)
		return 0;

	new->type = MPSS_FILE;
	if ((new->source = malloc(strlen(source) + 1)) == NULL) {
		free(new);
		return 0;
	}
	strcpy(new->source, source);

#ifdef SUPPORT_FILELIST
	new->uid = -1;
#endif

	insert_list(id, &dest[1], new, top, NULL, perrs);
	return 0;
}

int
_handle_rpm(int id, struct filetree *top, char *source, struct mpss_elist *perrs)
{
	struct stat sbuf;
	struct filetree *new;
	char tmp[PATH_MAX];
	char dest[PATH_MAX];

	if (lstat(source, &sbuf) != 0)
		return 1;

	if (!S_ISREG(sbuf.st_mode)) {
		add_perr(perrs, PERROR, "mic%d RPM '%s' is not a file\n", id, source);
		return 1;
	}

	if ((new = malloc(sizeof(struct filetree))) == NULL)
		return 1;

	new->type = MPSS_FILE;
	if ((new->source = malloc(strlen(source) + 1)) == NULL) {
		free(new);
		return 1;
	}
	strcpy(new->source, source);

#ifdef SUPPORT_FILELIST
	new->uid = -1;
#endif
	strncpy(tmp, source, PATH_MAX) ;
	snprintf(dest, PATH_MAX, "RPMs-to-install/%s", basename(source));
	insert_list(id, dest, new, top, NULL, perrs);
	return 0;
}

int
handle_rpm(int id, struct filetree *top, char *source, struct mpss_elist *perrs)
{
	struct stat sbuf;
	char path[PATH_MAX];
	char base[PATH_MAX];
	char *slash;
	struct dirent *file;
	DIR *dp;

	if (*source != '/') {
		add_perr(perrs, PERROR, "mic%d Failed RPM '%s' must start from the '/' directory\n", id, source);
		return 1;
	}

	strncpy(base, source, PATH_MAX);

	if (lstat(source, &sbuf) == 0) {
		if (S_ISDIR(sbuf.st_mode))
			snprintf(base, PATH_MAX, "%s/*.*", source);
	}

	if ((slash = strrchr(base, '/')) == NULL) {
		add_perr(perrs, PERROR, "mic%d Failed RPM '%s' must start from the '/' directory\n", id, source);
		return 1;
	}

	*slash = '\0';
	slash++;

	if ((dp = opendir(base)) == NULL) {
		add_perr(perrs, PERROR, "mic%d Failed to find RPM overlay '%s'\n", id, source);
		return 1;
	}

	while ((file = readdir(dp)) != NULL) {
		if (!strcmp(file->d_name, ".") ||
	 	    !strcmp(file->d_name, ".."))
			continue;

		if (fnmatch(slash, file->d_name, 0) == 0) {
			snprintf(path, PATH_MAX, "%s/%s", base, file->d_name);
			_handle_rpm(id, top, path, perrs);
		}
	}

	closedir(dp);
	return 0;
}

int
handle_subdir(int id, struct filetree *top, char *source, char *target, char *lead, struct mpss_elist *perrs)
{
	struct stat sbuf;

	if (lstat(source, &sbuf) != 0)
		return 0;

	if (S_ISREG(sbuf.st_mode))
		add_simple_file(id, source, target, lead, top, perrs);
	else if (S_ISCHR(sbuf.st_mode))
		add_simple_nod(id, 'c', source, target, lead, top, perrs);
	else if (S_ISBLK(sbuf.st_mode))
		add_simple_nod(id, 'b', source, target, lead, top, perrs);
	else if (S_ISFIFO(sbuf.st_mode))
		add_simple_pipe(id, source, target, lead, top, perrs);
	else if (S_ISLNK(sbuf.st_mode))
		add_simple_link(id, source, target, lead, top, perrs);
	else if (S_ISSOCK(sbuf.st_mode))
		add_simple_sock(id, source, target, lead, top, perrs);
	else if (S_ISDIR(sbuf.st_mode))
		add_simple_dir(id, source, target, lead, top, perrs);

	return 0;
}

int
handle_simple(int id, struct filetree *top, char *source, char *target, char *lead, struct mpss_elist *perrs)
{
	char path[PATH_MAX];
	char base[PATH_MAX];
	char *slash;
	struct dirent *file;
	DIR *dp;

	if (*source != '/') {
		add_perr(perrs, PERROR, "mic%d Failed overlay '%s' must start from the '/' directory\n", id, source);
		return 1;
	}

	strncpy(base, source, PATH_MAX);

	while (strlen(base) && (base[strlen(base) - 1] == '/'))
		base[strlen(base) - 1] = '\0';

	if ((slash = strrchr(base, '/')) == NULL) {
		add_perr(perrs, PERROR, "mic%d Failed RPM '%s' must start from the '/' directory\n", id, source);
		return 1;
	}

	*slash = '\0';
	slash++;

	if ((dp = opendir(base)) == NULL) {
		add_perr(perrs, PERROR, "mic%d Failed to find RPM overlay '%s'\n", id, source);
		return 1;
	}

	while ((file = readdir(dp)) != NULL) {
		if (!strcmp(file->d_name, ".") ||
	 	    !strcmp(file->d_name, ".."))
			continue;

		if (fnmatch(slash, file->d_name, 0) == 0) {
			snprintf(path, PATH_MAX, "%s/%s", base, file->d_name);
			handle_subdir(id, top, path, target, lead, perrs);
		}
	}

	closedir(dp);
	return 0;
}

#ifdef MPSSDEBUG
void
dump_tree(struct filetree *elem, int level)
{
	int i;

	while (elem) {
		for (i = 0; i < level; i++) {
			printf("     ");
		}

		if (elem->type == MPSS_DIR) {
			printf("| %s <==> %s\n", elem->name, elem->source);
			for (i = 0; i < level; i++) {
				printf("     ");
			}
			printf("\\----\\\n");
			dump_tree(elem->subdir, level + 1);
		} else if (elem->type == MPSS_LINK) {
			printf("| %s <==> %s\n", elem->name, elem->source);
		} else if (elem->type == MPSS_DNODE) {
			printf("| %s <==> %s\n", elem->name, elem->source);
		} else if (elem->type == MPSS_PIPE) {
			printf("| %s <==> %s\n", elem->name, elem->source);
		} else if (elem->type == MPSS_SOCK) {
			printf("| %s <==> %s\n", elem->name, elem->source);
		} else {
			printf("| %s <==> %s\n", elem->name, elem->source);
		}

		elem = elem->next;
	}
}
#endif

void
free_list(struct filetree *elem)
{
	struct filetree *freethis;

	while (elem) {
		if (elem->type == MPSS_DIR) {
			free_list(elem->subdir);
		}

		freethis = elem;
		elem = elem->next;
		free(freethis);
	}
}

int
handle_common(struct mpss_env *menv, int id, int type, char *dir, char *list, struct filetree *top, struct mpss_elist *perrs)
{
	char comname[PATH_MAX];
	struct dirent *file;
	DIR *dp;

	switch (type) {
	case SRCTYPE_DIR:
		mpssut_filename(menv, NULL, comname, PATH_MAX, "%s", dir);
		if ((dp = opendir(comname))) {
			while ((file = readdir(dp)) != NULL) {
				if (!strcmp(file->d_name, ".") ||
			 	    !strcmp(file->d_name, ".."))
					continue;

				mpssut_filename(menv, NULL, comname, PATH_MAX, "%s/%s", dir, file->d_name);
				handle_subdir(id, top, comname, "/", "", perrs);
			}

			closedir(dp);
		}

		break;
	case SRCTYPE_FILELIST:
		if (handle_dir(id, top, dir, list, perrs)) {
			return 3;
		}
		break;
	}

	return 0;
}

int
gen_fs_tree(struct mpss_env *menv, struct mic_info *mic, struct filetree *top, struct mpss_elist *perrs)
{
	struct moverdir *dir = mic->config.filesrc.overlays.next;

	top->type = MPSS_DIR;
	if ((top->name = malloc(2)) == NULL) {
		return 1;
	}
	strcpy(top->name, "/");
	top->source = NULL;
	top->mode = 0555;
	top->uid = 0;
	top->gid = 0;
	top->subdir = NULL;
	top->next = NULL;
	top->prev = NULL;

	if (mic->config.filesrc.base.type == SRCTYPE_DIR) {
		handle_common(menv, mic->id, mic->config.filesrc.base.type, mic->config.filesrc.base.dir,
			      mic->config.filesrc.base.list, top, perrs);
	}

	handle_common(menv, mic->id, mic->config.filesrc.common.type, mic->config.filesrc.common.dir,
		      mic->config.filesrc.common.list, top, perrs);

	while (dir) {
		if (dir->state == FALSE)
			goto nextdir;

		switch (dir->type) {
		case OVER_TYPE_SIMPLE:
			handle_simple(mic->id, top, dir->dir, dir->target, "", perrs);
			break;
		case OVER_TYPE_FILELIST:
			handle_dir(mic->id, top, dir->dir, dir->target, perrs);
			break;
		case OVER_TYPE_FILE:
			handle_file(mic->id, top, dir->dir, dir->target, perrs);
			break;
		case OVER_TYPE_RPM:
			handle_rpm(mic->id, top, dir->dir, perrs);
			break;
		}
nextdir:
		dir = dir->next;
	}

	handle_common(menv, mic->id, mic->config.filesrc.mic.type, mic->config.filesrc.mic.dir,
		      mic->config.filesrc.mic.list, top, perrs);

	return 0;
}

int
mpss_gen_initrd(struct mpss_env *menv, struct mic_info *mic, struct mpss_elist *perrs)
{
	struct filetree top;
	int err;

	mpss_clear_elist(perrs);


	if ((err = gen_fs_tree(menv, mic, &top, perrs)) == 0) {
#ifdef MPSSDEBUG
		dump_tree(&top, 0);
#endif
		err = gen_cpio(menv, mic, &top, perrs);
	}

	free_list(top.subdir);
	return err;
}

int
mpss_gen_nfsdir(struct mpss_env *menv, struct mic_info *mic, int createusr, struct mpss_elist *perrs)
{
	struct filetree top;
	int err;

	mpss_clear_elist(perrs);

	if ((err = gen_fs_tree(menv, mic, &top, perrs)) == 0) {
#ifdef MPSSDEBUG
		dump_tree(&top, 0);
#endif
		err = copy_nfs(menv, mic, &top, createusr, perrs);
	}

	free_list(top.subdir);
	return err;
}

void
insert_list(int id, char *path, struct filetree *new, struct filetree *elem, char *source, struct mpss_elist *perrs)
{
	struct filetree *tmp;
	char *cur = path;
	char *next = cur;
	int last = 0;
	int miss;

	if (strlen(path) == 0)
		return;

	while ((*next != '\0') && (*next != '/'))
		next++;

	if (*next == '\0')
		last = 1;

	*next = '\0';
	if (elem->subdir == NULL) {
		if (last) {
			elem->subdir = new;
			if ((new->name = malloc(strlen(cur) + 1)) == NULL) {
				return;
			}
			strcpy(new->name, cur);
			new->next = NULL;
			new->prev = NULL;
			return;
		} else {
			if ((elem->subdir = malloc(sizeof(struct filetree))) == NULL) {
				return;
			}
			elem = elem->subdir;
			elem->next = NULL;
			elem->prev = NULL;
			if ((elem->name = malloc(strlen(cur) + 1)) == NULL) {
				return;
			}
			strcpy(elem->name, cur);
			goto add_missing_entry;
		}
	}

	if (strcmp(cur, elem->subdir->name) < 0) {
		if (last) {
			new->next = elem->subdir;
			new->prev = NULL;
			elem->subdir->prev = new;
			elem->subdir = new;
			if ((new->name = malloc(strlen(cur) + 1)) == NULL) {
				return;
			}
			strcpy(new->name, cur);
			return;
		} else {
			if ((tmp = malloc(sizeof(struct filetree))) == NULL) {
				return;
			}
			tmp->next = elem->subdir;
			tmp->prev = NULL;
			elem->subdir->prev = tmp;
			elem->subdir = tmp;
			elem = tmp;
			//elem->next = NULL;
			goto add_missing_entry;
		}
	}

	elem = elem->subdir;

	while (elem != NULL) {
		miss = strcmp(cur, elem->name);
		if (miss < 0) {	// Insert before the cur element
			goto insert_here;
		} else if (miss == 0) {	// Exact match
			goto do_match;
		} else {		// Belongs after cur element
			tmp = elem;
			elem = elem->next;
		}
	}

	// Add to tail
	if (last) {
		tmp->next = new;
		new->prev = tmp;
		new->next = NULL;
		if ((new->name = malloc(strlen(cur) + 1)) == NULL) {
			return;
		}
		strcpy(new->name, cur);
		return;
	}

	// The entry has a directory in the middle that has not been previously identified.
	if ((tmp->next = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	tmp->next->prev = tmp;
	elem = tmp->next;
	elem->next = NULL;
	goto add_missing_entry;

insert_here:
	// End of line so just add the new piece
	if (last) {
		new->prev = elem->prev;
		new->next = elem;
		elem->prev->next = new;
		elem->prev = new;
		if ((new->name = malloc(strlen(cur) + 1)) == NULL) {
			return;
		}
		strcpy(new->name, cur);
		return;
	}

	// The entry has a directory in the middle that has not been previously identified.
	if ((tmp = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}
	tmp->prev = elem->prev;
	tmp->next = elem;
	elem->prev->next = tmp;
	elem->prev = tmp;
	elem = tmp;

add_missing_entry:
	elem->type = MPSS_DIR;
	if ((elem->name = malloc(strlen(cur) + 1)) == NULL) {
		free(elem);
		return;
	}
	strcpy(elem->name, cur);
	elem->source = NULL;
	elem->mode = 0755;
	elem->uid = 0;
	elem->gid = 0;
	elem->subdir = NULL;

do_match:
	if (!last) {
		// Recurse to next level in path
		insert_list(id, &next[1], new, elem, source, perrs);
		return;
	}

	// Entry types have to match
	if (elem->type != new->type) {
		add_perr(perrs, PINFO, "mic%d: overlay %s mismatched types changing from %s to %s",
			      id, path, get_type_string(elem->type), get_type_string(new->type));
		return;
	}

	// Update the old entry to match and free the new.
	if (elem->source != NULL)
		free(elem->source);

	elem->source = new->source;
	elem->mode = new->mode;
	elem->uid = new->uid;
	elem->gid = new->gid;
	free(new);
	return;
}

void
add_dir(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs)
{
	struct filetree *new;
	char path[PATH_MAX];

	if ((new = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	if (sscanf(line, "%" str(PATH_MAX) "s %o %d %d", path, &new->mode, &new->uid, &new->gid) != 4) {
		add_perr(perrs, PWARN, "mic%d Bad attribute line 'dir %s'", id, line);
	}

	// Check for trailing / characters
	while (path[strlen(path) - 1] == '/')
		path[strlen(path) - 1] = '\0';

	new->type = MPSS_DIR;
	new->source = NULL;
	new->subdir = NULL;
	new->next = NULL;
	new->prev = NULL;

	insert_list(id, &path[1], new, top, NULL, perrs);
}

void
simple_path(char *buf, int size, char *target, char *lead, char *filename)
{
	buf[0] = '\0';

	if (strcmp(target, "/")) {
		strcat(buf, target);
	}

	strcat(buf, "/");

	if (strlen(lead)) {
		strcat(buf, lead);
		strcat(buf, "/");
	}

	strcat(buf, filename);

}

void
add_simple_dir(int id, char *source, char *target, char *lead, struct filetree *top, struct mpss_elist *perrs)
{
	char newsource[PATH_MAX];
	char newlead[PATH_MAX];
	struct filetree *new;
	char path[PATH_MAX];
	char *filename = basename(source);
	struct dirent *file;
	DIR *dp;

	if ((new = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	new->type = MPSS_DIR;
	new->subdir = NULL;
	new->next = NULL;
	new->prev = NULL;

	if ((new->source = malloc(strlen(source) + 1)) == NULL) {
		free(new);
		return;
	}
	strcpy(new->source, source);

#ifdef SUPPORT_FILELIST
	new->uid = -1;
#endif
	simple_path(path, PATH_MAX, target, lead, filename);
	insert_list(id, &path[1], new, top, NULL, perrs);

	if ((dp = opendir(source)) == NULL) {
		return;
	}

	while ((file = readdir(dp)) != NULL) {
		if (!strcmp(file->d_name, ".") ||
		    !strcmp(file->d_name, "..")) {
			continue;
		}

		snprintf(newsource, PATH_MAX - 1, "%s/%s", source, file->d_name);
		if (strlen(lead))
			snprintf(newlead, PATH_MAX - 1, "%s/%s", lead, basename(source));
		else
			snprintf(newlead, PATH_MAX - 1, "%s", basename(source));
		handle_subdir(id, top, newsource, target, newlead, perrs);
	}

	closedir(dp);
}

void
add_file(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs)
{
	struct filetree *new;
	char path[PATH_MAX];
	char loc[PATH_MAX];

	if ((new = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	if (sscanf(line, "%" str(PATH_MAX) "s %" str(PATH_MAX) "s %o %d %d",
		   path, loc, &new->mode, &new->uid, &new->gid) != 5) {
		add_perr(perrs, PWARN, "mic%d Bad attribute line 'dir %s'", id, line);
	}

	new->type = MPSS_FILE;
	new->subdir = NULL;
	new->next = NULL;

	if ((new->source = malloc(strlen(sourcedir) + strlen(loc) + 3)) == NULL) {
		free(new);
		return;
	}
	sprintf(new->source, "%s/%s", sourcedir, loc);

	insert_list(id, &path[1], new, top, NULL, perrs);
}

void
add_simple_file(int id, char *source, char *target, char *lead, struct filetree *top, struct mpss_elist *perrs)
{
	struct filetree *new;
	char path[PATH_MAX];
	char *filename = basename(source);

	if ((new = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	new->type = MPSS_FILE;

	if ((new->source = malloc(strlen(source) + 1)) == NULL) {
		free(new);
		return;
	}
	strcpy(new->source, source);

#ifdef SUPPORT_FILELIST
	new->uid = -1;
#endif

	simple_path(path, PATH_MAX, target, lead, filename);
	new->subdir = NULL;
	new->next = NULL;

	insert_list(id, &path[1], new, top, NULL, perrs);
}

void
add_link(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs)
{
	struct filetree *new;
	char path[PATH_MAX];
	char target[PATH_MAX];

	if ((new = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	if (sscanf(line, "%" str(PATH_MAX) "s %" str(PATH_MAX) "s %o %d %d",
		   path, target, &new->mode, &new->uid, &new->gid) != 5) {
		add_perr(perrs, PWARN, "mic%d Bad attribute line 'slink %s'", id, line);
	}

	new->type = MPSS_LINK;

	if ((new->link = malloc(strlen(target) + 1)) == NULL) {
		free(new);
		return;
	}

	strcpy(new->link, target);
	new->subdir = NULL;
	new->next = NULL;

	insert_list(id, &path[1], new, top, NULL, perrs);
}

void
add_simple_link(int id, char *source, char *target, char *lead, struct filetree *top, struct mpss_elist *perrs)
{
	struct filetree *new;
	char path[PATH_MAX];
	char rbuf[PATH_MAX];
	char *filename = basename(source);

	if ((new = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	memset(new, 0, sizeof(struct filetree));
	new->type = MPSS_LINK;

	if ((new->source = malloc(strlen(source) + 1)) == NULL) {
		free(new);
		return;
	}
	strcpy(new->source, source);

	memset(rbuf, 0, PATH_MAX);
	readlink(source, rbuf, PATH_MAX);
	if ((new->link = malloc(strlen(rbuf) + 1)) == NULL) {
		free(new->source);
		free(new);
		return;
	}
	memset(new->link, 0, strlen(rbuf) + 1);
	strcpy(new->link, rbuf);

#ifdef SUPPORT_FILELIST
	new->uid = -1;
#endif

	simple_path(path, PATH_MAX, target, lead, filename);
	new->subdir = NULL;
	new->next = NULL;

	insert_list(id, &path[1], new, top, NULL, perrs);
}

void
add_dnode(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs)
{
	struct filetree *new;
	char path[PATH_MAX];

	if ((new = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	if (sscanf(line, "%" str(PATH_MAX) "s %o %d %d %c %u %u", path, &new->mode, &new->uid,
		   &new->gid, &new->inode_type, &new->major, &new->minor) != 7) {
		add_perr(perrs, PWARN, "mic%d Bad attribute line 'dir %s'", id, line);
	}

	new->type = MPSS_DNODE;
	new->subdir = NULL;
	new->next = NULL;
	new->prev = NULL;

	insert_list(id, &path[1], new, top, NULL, perrs);
}

void
add_simple_nod(int id, char type, char *source, char *target, char *lead, struct filetree *top, struct mpss_elist *perrs)
{
	struct filetree *new;
	char path[PATH_MAX];
	char *filename = basename(source);

	if ((new = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	new->type = MPSS_DNODE;
	new->source = NULL;
	new->subdir = NULL;
	new->next = NULL;
	new->prev = NULL;

	if ((new->source = malloc(strlen(source) + 1)) == NULL) {
		free(new);
		return;
	}
	strcpy(new->source, source);

#ifdef SUPPORT_FILELIST
	new->uid = -1;
#endif
	new->inode_type = type;
	simple_path(path, PATH_MAX, target, lead, filename);

	insert_list(id, &path[1], new, top, NULL, perrs);
}

void
add_pipe(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs)
{
	struct filetree *new;
	char path[PATH_MAX];

	if ((new = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	if (sscanf(line, "%" str(PATH_MAX) "s %o %d %d", path, &new->mode, &new->uid, &new->gid) != 4) {
		add_perr(perrs, PWARN, "mic%d Bad attribute line 'dir %s'", id, line);
	}

	new->type = MPSS_PIPE;
	new->source = NULL;
	new->subdir = NULL;
	new->next = NULL;
	new->prev = NULL;

	insert_list(id, &path[1], new, top, NULL, perrs);
}

void
add_simple_pipe(int id, char *source, char *target, char *lead, struct filetree *top, struct mpss_elist *perrs)
{
	struct filetree *new;
	char path[PATH_MAX];
	char *filename = basename(source);

	if ((new = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	new->type = MPSS_PIPE;
	new->source = NULL;
	new->subdir = NULL;
	new->next = NULL;
	new->prev = NULL;

	if ((new->source = malloc(strlen(source) + 1)) == NULL) {
		free(new);
		return;
	}
	strcpy(new->source, source);

#ifdef SUPPORT_FILELIST
	new->uid = -1;
#endif
	simple_path(path, PATH_MAX, target, lead, filename);

	insert_list(id, &path[1], new, top, NULL, perrs);
}

void
add_sock(int id, char *line, struct filetree *top, char *sourcedir, struct mpss_elist *perrs)
{
	struct filetree *new;
	char path[PATH_MAX];

	if ((new = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	if (sscanf(line, "%" str(PATH_MAX) "s %o %d %d", path, &new->mode, &new->uid, &new->gid) != 4) {
		add_perr(perrs, PWARN, "mic%d Bad attribute line 'dir %s'", id, line);
	}

	new->type = MPSS_SOCK;
	new->source = NULL;
	new->subdir = NULL;
	new->next = NULL;
	new->prev = NULL;

	insert_list(id, &path[1], new, top, NULL, perrs);
}

void
add_simple_sock(int id, char *source, char *target, char *lead, struct filetree *top, struct mpss_elist *perrs)
{
	struct filetree *new;
	char path[PATH_MAX];
	char *filename = basename(source);

	if ((new = malloc(sizeof(struct filetree))) == NULL) {
		return;
	}

	new->type = MPSS_SOCK;
	new->source = NULL;
	new->subdir = NULL;
	new->next = NULL;
	new->prev = NULL;

	if ((new->source = malloc(strlen(source) + 1)) == NULL) {
		free(new);
		return;
	}
	strcpy(new->source, source);

#ifdef SUPPORT_FILELIST
	new->uid = -1;
#endif
	simple_path(path, PATH_MAX, target, lead, filename);

	insert_list(id, &path[1], new, top, NULL, perrs);
}

void
push_rest(char *name, FILE *fp, unsigned int *inode, unsigned int *offset)
{
	unsigned int name_len = strlen(name) + 1;
	unsigned int tmp;

	fputs(name, fp);
	fputc('\0', fp);
	*offset += name_len;

	tmp = name_len + 110;
	while (tmp & 3) {
		fputc('\0', fp);
		(*offset)++;
		tmp++;
	}
}

void
gen_dir(struct mpss_env *menv, struct filetree *d, char *leader, FILE *fp, unsigned int *inode, unsigned int *offset, struct mpss_elist *perrs)
{
	time_t mtime = time(NULL);
	unsigned int name_len;
	char name[PATH_MAX];
	struct stat sbuf;

	name_len = snprintf(name, PATH_MAX, "%s/%s", leader, d->name);

#ifdef SUPPORT_FILELIST
	if (d->uid != -1) {
		sbuf.st_uid = d->uid;
		sbuf.st_gid = d->gid;
		sbuf.st_mode = d->mode;
	} else {
#endif
	if (stat(d->source, &sbuf)) {
		add_perr(perrs, PERROR, "[GenDir] '/%s' Invalid source file %s: %s", d->name, d->source, strerror(errno));
		return;
	}
#ifdef SUPPORT_FILELIST
	}
#endif

	fprintf(fp,"%s%08X%08X%08lX%08lX%08X%08lX%08X%08X%08X%08X%08X%08X%08X",
		"070701",		/* magic */
		(*inode)++,		/* ino */
		(sbuf.st_mode & 0777) | S_IFDIR, /* mode */
		(long) sbuf.st_uid,	/* uid */
		(long) sbuf.st_gid,	/* gid */
		2,			/* nlink */
		(long) mtime,		/* mtime */
		0,			/* filesize */
		3,			/* major */
		1,			/* minor */
		0,			/* rmajor */
		0,			/* rminor */
		name_len,		/* namesize */
		0);			/* chksum */

	*offset += 110;
	push_rest(&name[1], fp, inode, offset);
	return;
}

void
gen_node(struct mpss_env *menv, struct filetree *n, char *leader, FILE *fp, unsigned int *inode, unsigned int *offset, struct mpss_elist *perrs)
{
	time_t mtime = time(NULL);
	unsigned int name_len;
	char name[PATH_MAX];
	struct stat sbuf;

	name_len = snprintf(name, PATH_MAX, "%s/%s", leader, n->name);

#ifdef SUPPORT_FILELIST
	if (n->uid != -1) {
		sbuf.st_uid = n->uid;
		sbuf.st_gid = n->gid;
		sbuf.st_mode = n->mode;
		sbuf.st_rdev = MKDEV(n->major, n->minor);
	} else {
#endif
	if (stat(n->source, &sbuf)) {
		add_perr(perrs, PERROR, "[GenNode] Invalid source file %s: %s", n->source, strerror(errno));
		return;
	}
#ifdef SUPPORT_FILELIST
	}
#endif

	if (n->inode_type == 'b')
		sbuf.st_mode = (sbuf.st_mode & 0777) | S_IFBLK;
	else
		sbuf.st_mode = (sbuf.st_mode & 0777) | S_IFCHR;

	fprintf(fp,"%s%08X%08X%08lX%08lX%08X%08lX%08X%08X%08X%08X%08X%08X%08X",
		"070701",		/* magic */
		(*inode)++,		/* ino */
		sbuf.st_mode,		/* mode */
		(long) sbuf.st_uid, 	/* uid */
		(long) sbuf.st_gid, 	/* gid */
		1,			/* nlink */
		(long) mtime,		/* mtime */
		0,			/* filesize */
		3,			/* major */
		1,			/* minor */
		(unsigned int)MAJOR(sbuf.st_rdev), /* rmajor */
		(unsigned int)MINOR(sbuf.st_rdev), /* rminor */
		name_len,		/* namesize */
		0);			/* chksum */

	*offset += 110;
	push_rest(&name[1], fp, inode, offset);
	return;
}

void
gen_pipe(struct mpss_env *menv, struct filetree *p, char *leader, FILE *fp, unsigned int *inode, unsigned int *offset, struct mpss_elist *perrs)
{
	time_t mtime = time(NULL);
	unsigned int name_len;
	char name[PATH_MAX];
	struct stat sbuf;

	name_len = snprintf(name, PATH_MAX, "%s/%s", leader, p->name);

#ifdef SUPPORT_FILELIST
	if (p->uid != -1) {
		sbuf.st_uid = p->uid;
		sbuf.st_gid = p->gid;
		sbuf.st_mode = p->mode;
	} else {
#endif
	if (stat(p->source, &sbuf)) {
		add_perr(perrs, PERROR, "[GenPipe] Invalid source file %s: %s", p->source, strerror(errno));
		return;
	}
#ifdef SUPPORT_FILELIST
	}
#endif


	fprintf(fp,"%s%08X%08X%08lX%08lX%08X%08lX%08X%08X%08X%08X%08X%08X%08X",
		"070701",		/* magic */
		(*inode)++,		/* ino */
		(sbuf.st_mode & 0777) | S_IFIFO, /* mode */
		(long) sbuf.st_uid, 	/* uid */
		(long) sbuf.st_gid, 	/* gid */
		2,			/* nlink */
		(long) mtime,		/* mtime */
		0,			/* filesize */
		3,			/* major */
		1,			/* minor */
		0,			/* rmajor */
		0,			/* rminor */
		name_len,		/* namesize */
		0);			/* chksum */

	*offset += 110;
	push_rest(&name[1], fp, inode, offset);
	return;
}

void
gen_sock(struct mpss_env *menv, struct filetree *s, char *leader, FILE *fp, unsigned int *inode, unsigned int *offset, struct mpss_elist *perrs)
{
	time_t mtime = time(NULL);
	unsigned int name_len;
	char name[PATH_MAX];
	struct stat sbuf;

	name_len = snprintf(name, PATH_MAX, "%s/%s", leader, s->name);

#ifdef SUPPORT_FILELIST
	if (s->uid != -1) {
		sbuf.st_uid = s->uid;
		sbuf.st_gid = s->gid;
		sbuf.st_mode = s->mode;
	} else {
#endif
	if (stat(s->source, &sbuf)) {
		add_perr(perrs, PERROR, "[GenSock] Invalid source file %s: %s", s->source, strerror(errno));
		return;
	}
#ifdef SUPPORT_FILELIST
	}
#endif

	fprintf(fp,"%s%08X%08X%08lX%08lX%08X%08lX%08X%08X%08X%08X%08X%08X%08X",
		"070701",		/* magic */
		(*inode)++,		/* ino */
		(sbuf.st_mode & 0777) | S_IFSOCK, /* mode */
		(long) sbuf.st_uid, 	/* uid */
		(long) sbuf.st_gid, 	/* gid */
		2,			/* nlink */
		(long) mtime,		/* mtime */
		0,			/* filesize */
		3,			/* major */
		1,			/* minor */
		0,			/* rmajor */
		0,			/* rminor */
		name_len,		/* namesize */
		0);			/* chksum */
	*offset += 110;
	push_rest(&name[1], fp, inode, offset);
	return;
}

void
gen_slink(struct mpss_env *menv, struct filetree *l, char *leader, FILE *fp, unsigned int *inode, unsigned int *offset, struct mpss_elist *perrs)
{
	unsigned int target_len = strlen(l->link) + 1;
	time_t mtime = time(NULL);
	unsigned int name_len;
	char name[PATH_MAX];
	struct stat sbuf;

	name_len = snprintf(name, PATH_MAX, "%s/%s", leader, l->name);

#ifdef SUPPORT_FILELIST
	if (l->uid != -1) {
		sbuf.st_uid = l->uid;
		sbuf.st_gid = l->gid;
		sbuf.st_mode = l->mode;
	} else {
#endif
	if (lstat(l->source, &sbuf)) {
		add_perr(perrs, PERROR, "[GenSlink] Invalid source file %s: %s", l->source, strerror(errno));
		return;
	}
#ifdef SUPPORT_FILELIST
	}
#endif

	fprintf(fp,"%s%08X%08X%08lX%08lX%08X%08lX%08X%08X%08X%08X%08X%08X%08X",
		"070701",		/* magic */
		(*inode)++,		/* ino */
		(sbuf.st_mode & 0777) | S_IFLNK, /* mode */
		(long) sbuf.st_uid, 	/* uid */
		(long) sbuf.st_gid,	/* gid */
		1,			/* nlink */
		(long) mtime,		/* mtime */
		target_len,		/* filesize */
		3,			/* major */
		1,			/* minor */
		0,			/* rmajor */
		0,			/* rminor */
		name_len,		/* namesize */
		0);			/* chksum */
	*offset += 110;

	fputs(&name[1], fp);
	fputc('\0', fp);
	*offset += name_len;

	while (*offset & 3) {
		fputc('\0', fp);
		(*offset)++;
	}

	fputs(l->link, fp);
	fputc('\0', fp);
	*offset += target_len;

	while (*offset & 3) {
		fputc('\0', fp);
		(*offset)++;
	}
}

void
gen_file(struct mpss_env *menv, struct filetree *f, char *leader, FILE *fp, unsigned int *inode, unsigned int *offset,
	 struct mpss_elist *perrs)
{
	char name[PATH_MAX];
	char src[PATH_MAX];
	struct stat sbuf;
	struct stat lbuf;
	time_t mtime = time(NULL);
	int fd = -1;
	char *filebuf = NULL;
	unsigned int name_len;
	ssize_t rsize;;

	snprintf(name, PATH_MAX, "%s/%s", leader, f->name);
	name_len = strlen(name);

	if (!mpssut_filename(menv, NULL, src, PATH_MAX, "%s", f->source)) {
		add_perr(perrs, PERROR, "Invalid source file: %s", f->source);
		return;
	}

	if ((fd = open(src, O_RDONLY)) < 0) {
		add_perr(perrs, PERROR, "Could not open source file: %s", f->source);
		return;
	}

	if (fstat(fd, &sbuf) != 0) {
		add_perr(perrs, PERROR, "CPIO gen %s fstat failed: %s", f->source, strerror(errno));
		close(fd);
		return;
	}

	if (lstat(src, &lbuf) != 0) {
		add_perr(perrs, PERROR, "CPIO gen %s lstat failed: %s", f->source, strerror(errno));
		close(fd);
		return;
	}

	if (S_ISLNK(lbuf.st_mode)) {
		add_perr(perrs, PERROR, "CPIO gen %s file has changed to a symbolic link: %s", f->source, strerror(errno));
		close(fd);
		return;
	}

	if (sbuf.st_size > ((unsigned int)0xffffffff - 1)) {
		add_perr(perrs, PERROR, "CPIO gen %s field width not sufficient for storing file size", f->source);
		close(fd);
		return;
	}

	if ((filebuf = malloc(sbuf.st_size)) == NULL) {
		add_perr(perrs, PERROR, "CPIO gen %s: malloc faileds", f->source);
		close(fd);
		return;
	}

	if ((rsize = read(fd, filebuf, sbuf.st_size)) != sbuf.st_size) {
		add_perr(perrs, PERROR, "CPIO gen %s: read size %d does not match stat size %d",
			 f->source, rsize, sbuf.st_size);
		free(filebuf);
		close(fd);
		return;
	}

#ifdef SUPPORT_FILELIST
	if (f->uid != -1) {
		sbuf.st_uid = f->uid;
		sbuf.st_gid = f->gid;
		sbuf.st_mode = f->mode;
	}
#endif

	fprintf(fp,"%s%08X%08X%08lX%08lX%08X%08lX%08lX%08X%08X%08X%08X%08X%08X",
		"070701",		/* magic */
		(*inode)++,		/* ino */
		(sbuf.st_mode & 0777) | S_IFREG, /* mode */
		(long) sbuf.st_uid, 	/* uid */
		(long) sbuf.st_gid, 	/* gid */
		1,			/* nlink */
		(long) mtime,		/* mtime */
		rsize,			/* filesize */
		3,			/* major */
		1,			/* minor */
		0,			/* rmajor */
		0,			/* rminor */
		name_len,		/* namesize */
		0);			/* chksum */
	*offset += 110;

	fputs(&name[1], fp);
	fputc('\0', fp);
	*offset += name_len;

	while(*offset & 3) {
		fputc('\0', fp);
		(*offset)++;
	}

	fwrite(filebuf, sbuf.st_size, 1, fp);
	*offset += sbuf.st_size;
	while(*offset & 3) {
		fputc('\0', fp);
		(*offset)++;
	}

	free(filebuf);
	close(fd);
}

void
follow_dir(struct mpss_env *menv, struct filetree *elem, char *leader, FILE *fp, unsigned int *inode, unsigned int *offset, struct mpss_elist *perrs)
{
	char *newleader;

	while (elem != NULL) {
		switch (elem->type) {
		case MPSS_DIR:
			gen_dir(menv, elem, leader, fp, inode, offset, perrs);
			if ((newleader = malloc(strlen(leader) + strlen(elem->name) + 2)) == NULL) {
				return;
			}
			sprintf(newleader, "%s/%s", leader, elem->name);
			follow_dir(menv, elem->subdir, newleader, fp, inode, offset, perrs);
			free(newleader);
			break;

		case MPSS_FILE:
			gen_file(menv, elem, leader, fp, inode, offset, perrs);
			break;

		case MPSS_LINK:
			gen_slink(menv, elem, leader, fp, inode, offset, perrs);
			break;

		case MPSS_DNODE:
			gen_node(menv, elem, leader, fp, inode, offset, perrs);
			break;

		case MPSS_PIPE:
			gen_pipe(menv, elem, leader, fp, inode, offset, perrs);
			break;

		case MPSS_SOCK:
			gen_sock(menv, elem, leader, fp, inode, offset, perrs);
			break;
		}

		elem = elem->next;
	}
}

void
cpio_trailer(FILE *fp, unsigned int *inode, unsigned int *offset)
{
	char *name = "TRAILER!!!";
	unsigned int name_len = strlen(name) + 1;

	fprintf(fp,"%s%08X%08X%08lX%08lX%08X%08lX%08X%08X%08X%08X%08X%08X%08X",
		"070701",		/* magic */
		0,			/* ino */
		0,			/* mode */
		(long) 0,		/* uid */
		(long) 0,		/* gid */
		1,			/* nlink */
		(long) 0,		/* mtime */
		0,			/* filesize */
		0,			/* major */
		0,			/* minor */
		0,			/* rmajor */
		0,			/* rminor */
		name_len,		/* namesize */
		0);			/* chksum */
	*offset += 110;
	push_rest(name, fp, inode, offset);

	while(*offset % 512) {
		fputc('\0', fp);
		(*offset)++;
	}
}

void
gzip_image(char *name)
{
	pid_t pid;
	char *ifargv[3];

	pid = fork();
	if (pid == 0) {
		ifargv[0] = "/bin/gzip";
		ifargv[1] = name;
		ifargv[2] = NULL;
		execve("/bin/gzip", ifargv, NULL);
	}

	waitpid(pid, NULL, 0);
}

void
gunzip_image(char *name)
{
	pid_t pid;
	char *ifargv[3];

	pid = fork();
	if (pid == 0) {
		ifargv[0] = "/bin/gunzip";
		ifargv[1] = name;
		ifargv[2] = NULL;
		execve("/bin/gunzip", ifargv, NULL);
	}

	waitpid(pid, NULL, 0);
}

int
mpss_cpioin(char *cfile, char *cdir, int dousr)
{
	pid_t pid;
	char *ifargv[6];
	int status;

	pid = fork();
	if (pid == 0) {

		if (chdir(cdir) < 0)
			exit(errno);

		fclose(stdout);
		fclose(stderr);
		ifargv[0] = "/bin/cpio";
		ifargv[1] = "-i";
		ifargv[2] = "-F";
		ifargv[3] = cfile;
		if (dousr) {
			ifargv[4] = "u*";
			ifargv[5] = NULL;
		} else {
			ifargv[4] = NULL;
		}
		execve("/bin/cpio", ifargv, NULL);
	}

	if (waitpid(pid, &status, 0) < 0)
		return status;

	return 0;
}

int
gen_cpio(struct mpss_env *menv, struct mic_info *mic, struct filetree *top, struct mpss_elist *perrs)
{
	char gzname[PATH_MAX];
	char cpioname[PATH_MAX];
	FILE *cpiofp;
	mode_t oldmask;
	struct stat sbuf;
	unsigned int inode = 721;  // simply somewhere to start
	unsigned int offset = 0;

	if (mpssut_filename(menv, NULL, gzname, PATH_MAX, "%s", mic->config.rootdev.target))
		unlink(gzname);

	if (strcmp(&gzname[strlen(gzname) - 3], ".gz")) {
		add_perr(perrs, PERROR, "mic%d Image file '%s' must end in '.gz'",
				mic->name, mic->config.rootdev.target);
		return 1;
	}

	strcpy(cpioname, gzname);
	cpioname[strlen(cpioname) - 3] = '\0';
	if (stat(cpioname, &sbuf) == 0)
		unlink(cpioname);

	if (mic->config.filesrc.base.type == SRCTYPE_CPIO) {
		unzip_base_cpio(mic->name, mic->config.filesrc.base.image, gzname, perrs);
	}

	oldmask = umask(S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
	if ((cpiofp = fopen(cpioname, "a")) == NULL) {
		umask(oldmask);
		return errno;
	}

	follow_dir(menv, top->subdir, "", cpiofp, &inode, &offset, perrs);
	cpio_trailer(cpiofp, &inode, &offset);

	fclose(cpiofp);
	umask(oldmask);
	gzip_image(cpioname);
	return 0;
}

void
copy_file(struct mpss_env *menv, struct filetree *f, char *leader, struct mpss_elist *perrs)
{
	struct stat sbuf;
	char to[PATH_MAX];
	char *buffer;
	int rsize;
	int fd;

	if (mpssut_filename(menv, NULL, to, PATH_MAX, "%s/%s", leader, f->name))
		unlink(to);

	if ((fd = open(f->source, O_RDONLY)) < 0) {
		add_perr(perrs, PERROR, "Cannot open source file %s: %s", f->source, strerror(errno));
		return;
	}

	if (fstat(fd, &sbuf) != 0) {
		add_perr(perrs, PERROR, "Cannot find source file %s: %s", f->source, strerror(errno));
		close(fd);
		return;
	}

#ifdef SUPPORT_FILELIST
	if (f->uid != -1) {
		sbuf.st_uid = f->uid;
		sbuf.st_gid = f->gid;
		sbuf.st_mode = f->mode;
	}
#endif

	if ((buffer = malloc(sbuf.st_size + 1)) == NULL) {
		add_perr(perrs, PERROR, "Memory allocation error: %s", strerror(errno));
		close(fd);
		return;
	}

	rsize = read(fd, buffer, sbuf.st_size);
	close(fd);

	if ((fd = open(to, O_WRONLY|O_CREAT, sbuf.st_mode & 0777)) < 0) {
		add_perr(perrs, PERROR, "Cannot open destination file %s: %s", to, strerror(errno));
		free(buffer);
		return;
	}

	write(fd, buffer, rsize);
	fchown(fd, sbuf.st_uid, sbuf.st_gid);
	close(fd);
	free(buffer);
}

void
copy_slink(struct mpss_env *menv, struct filetree *l, char *leader, struct mpss_elist *perrs)
{
	char name[PATH_MAX];
	struct stat sbuf;

	if (mpssut_filename(menv, NULL, name, PATH_MAX, "%s/%s", leader, l->name))
		unlink(name);

#ifdef SUPPORT_FILELIST
	if (l->uid != -1) {
		sbuf.st_uid = l->uid;
		sbuf.st_gid = l->gid;
		sbuf.st_mode = l->mode;
	} else {
#endif
	if (lstat(l->source, &sbuf)) {
		add_perr(perrs, PERROR, "[CopySlink] Invalid source file %s: %s", l->source, strerror(errno));
		return;
	}
#ifdef SUPPORT_FILELIST
	}
#endif

	if (!symlink(l->link, name)) {
		chown(name, sbuf.st_uid, sbuf.st_gid);
		chmod(name, sbuf.st_mode & 0777);
	} else {
		add_perr(perrs, PERROR, "Error making symlink %s: %d", name, errno);
	}
}

void
copy_node(struct mpss_env *menv, struct filetree *n, char *leader, struct mpss_elist *perrs)
{
	char name[PATH_MAX];
	struct stat sbuf;

	if (mpssut_filename(menv, NULL, name, PATH_MAX, "%s/%s", leader, n->name))
		unlink(name);

#ifdef SUPPORT_FILELIST
	if (n->uid != -1) {
		sbuf.st_uid = n->uid;
		sbuf.st_gid = n->gid;
		sbuf.st_mode = n->mode;
		sbuf.st_dev = MKDEV(n->major, n->minor);
	} else {
#endif
	if (stat(n->source, &sbuf)) {
		add_perr(perrs, PERROR, "[CopyNode] Invalid source file %s: %s", n->source, strerror(errno));
		return;
	}
#ifdef SUPPORT_FILELIST
	}
#endif

	if (n->inode_type == 'b')
		sbuf.st_mode = (sbuf.st_mode & 0777) | S_IFBLK;
	else
		sbuf.st_mode = (sbuf.st_mode & 0777) | S_IFCHR;

	if (!mknod(name, sbuf.st_mode, sbuf.st_dev))
		chown(name, sbuf.st_uid, sbuf.st_gid);
	else
		add_perr(perrs, PERROR, "Error making node %s: %s", name, strerror(errno));
}

void
copy_pipe(struct mpss_env *menv, struct filetree *p, char *leader, struct mpss_elist *perrs)
{
	char name[PATH_MAX];
	struct stat sbuf;

	if (mpssut_filename(menv, NULL, name, PATH_MAX, "%s/%s", leader, p->name))
		unlink(name);

#ifdef SUPPORT_FILELIST
	if (p->uid != -1) {
		sbuf.st_uid = p->uid;
		sbuf.st_gid = p->gid;
		sbuf.st_mode = p->mode;
	} else {
#endif
	if (stat(p->source, &sbuf)) {
		add_perr(perrs, PERROR, "[CopyPipe] Invalid source file %s: %s", p->source, strerror(errno));
		return;
	}
#ifdef SUPPORT_FILELIST
	}
#endif

	if (!mknod(name, (sbuf.st_mode & 0777) | S_IFIFO, 0))
		chown(name, sbuf.st_uid, sbuf.st_gid);
	else
		add_perr(perrs, PERROR, "Error making pipe %s: %d", name, errno);
}

void
copy_sock(struct mpss_env *menv, struct filetree *s, char *leader, struct mpss_elist *perrs)
{
	char name[PATH_MAX];
	struct stat sbuf;

	if (mpssut_filename(menv, NULL, name, PATH_MAX, "%s/%s", leader, s->name))
		unlink(name);

#ifdef SUPPORT_FILELIST
	if (s->uid != -1) {
		sbuf.st_uid = s->uid;
		sbuf.st_gid = s->gid;
		sbuf.st_mode = s->mode;
	} else {
#endif
	if (stat(s->source, &sbuf)) {
		add_perr(perrs, PERROR, "[CopySock] Invalid source file %s: %s", s->source, strerror(errno));
		return;
	}
#ifdef SUPPORT_FILELIST
	}
#endif

	if (!mknod(name, (sbuf.st_mode & 0777) | S_IFSOCK, 0))
		chown(name, sbuf.st_uid, sbuf.st_gid);
	else
		add_perr(perrs, PERROR, "Error making node %s: %d", name, errno);
}

void
copy_dir(struct mpss_env *menv, struct filetree *elem, char *leader, int level, struct mpss_elist *perrs)
{
	char name[PATH_MAX];
	char *newleader;
	struct stat sbuf;

	while (elem != NULL) {
		switch (elem->type) {
		case MPSS_DIR:
			if (!mpssut_filename(menv, NULL, name, PATH_MAX, "%s/%s", leader, elem->name)) {
#ifdef SUPPORT_FILELIST
				if (elem->uid != -1) {
					sbuf.st_uid = elem->uid;
					sbuf.st_gid = elem->gid;
					sbuf.st_mode = elem->mode;
				} else {
#endif
				if (stat(elem->source, &sbuf)) {
					add_perr(perrs, PERROR, "[CopyDir] Invalid source file: %s", elem->source);
					return;
				}
#ifdef SUPPORT_FILELIST
				}
#endif
				if (!mkdir(name, sbuf.st_mode & 0777)) {
					chown(name, sbuf.st_uid, sbuf.st_gid);
				} else {
					add_perr(perrs, PERROR, "[CopyDir] Error creating dir %s: %s", name, strerror(errno));
					return;
				}
			}

			if ((newleader = malloc(strlen(leader) + strlen(elem->name) + 2)) == NULL)
				return;

			sprintf(newleader, "%s/%s", leader, elem->name);
			copy_dir(menv, elem->subdir, newleader, level + 1, perrs);
			free(newleader);
			break;

		case MPSS_FILE:
			copy_file(menv, elem, leader, perrs);
			break;

		case MPSS_LINK:
			copy_slink(menv, elem, leader, perrs);
			break;

		case MPSS_DNODE:
			copy_node(menv, elem, leader, perrs);
			break;

		case MPSS_PIPE:
			copy_pipe(menv, elem, leader, perrs);
			break;

		case MPSS_SOCK:
			copy_sock(menv, elem, leader, perrs);
			break;
		}

		elem = elem->next;
	}
}

int
copy_nfs(struct mpss_env *menv, struct mic_info *mic, struct filetree *top, int createusr, struct mpss_elist *perrs)
{
	char src[PATH_MAX];
	char orig[PATH_MAX];
	char dest[PATH_MAX];
	char usrbase[PATH_MAX];
	char *lastslash;
	char *newleader;
	struct stat sbuf;
	struct filetree *elem = top->subdir;
	char *leader = NULL;

	if (mic->config.filesrc.base.type == SRCTYPE_CPIO) {
		mpssut_filename(menv, NULL, dest, PATH_MAX, "%s.%d.gz",
			      strchr(mic->config.rootdev.target, ':') + 1, getpid());
		mpssut_filename(menv, NULL, orig, PATH_MAX, "%s", mic->config.filesrc.base.image);
		mpssut_filename(menv, NULL, src, PATH_MAX, "%s.%d",
			      strchr(mic->config.rootdev.target, ':') + 1, getpid());
		unzip_base_cpio(mic->name, orig, dest, perrs);

		mpssut_filename(menv, NULL, dest, PATH_MAX, "%s", strchr(mic->config.rootdev.target, ':') + 1);

		if (createusr) {
			mpssut_filename(menv, NULL, usrbase, PATH_MAX, "%s", strchr(mic->config.rootdev.nfsusr, ':') + 1);
			if ((lastslash = strrchr(dest, '/')) != NULL)
				*lastslash = '\0';
			mpss_cpioin(src, dest, TRUE);
			strcat(dest, "/usr");
			rename(dest, usrbase);
		} else {
			mpssut_filename(menv, NULL, dest, PATH_MAX, "%s", strchr(mic->config.rootdev.target, ':') + 1);
			mpss_cpioin(src, dest, FALSE);
		}

		unlink(src);
	}

	while (elem != NULL) {
		if (createusr) {
			if ((mic->config.rootdev.type == ROOT_TYPE_SPLITNFS) &&
			    !strcmp(elem->name, "usr")) {
				leader = strchr(mic->config.rootdev.nfsusr, ':') + 1;
				copy_dir(menv, elem->subdir, leader, 1, perrs);
				return 0;
			} else {
				elem = elem->next;
				continue;
			}
		} else {
			if ((mic->config.rootdev.type == ROOT_TYPE_SPLITNFS) &&
			    !strcmp(elem->name, "usr") && leader) {
				if (!mpssut_filename(menv, &sbuf, src, PATH_MAX, "%s/usr", leader)) {
					if (!mkdir(src, elem->mode)) {
#ifdef SUPPORT_FILELIST
						if (elem->uid != -1) {
							sbuf.st_uid = elem->uid;
							sbuf.st_gid = elem->gid;
							sbuf.st_mode = elem->mode;
						} else {
#endif
						if (stat(elem->source, &sbuf)) {
							add_perr(perrs, PERROR, "[CopyDir] Invalid source file: %s",
									elem->source);
							return 1;
						}
#ifdef SUPPORT_FILELIST
						}
#endif

						chown(src, sbuf.st_uid, sbuf.st_gid);
						chmod(src, sbuf.st_mode & 0777);
					} else {
						add_perr(perrs, PERROR, "Error creating dir %s: %d", src, errno);
					}
				}


				elem = elem->next;
				continue;
			} else {
				leader = strchr(mic->config.rootdev.target, ':') + 1;
			}
		}

		switch (elem->type) {
		case MPSS_DIR:
			mpssut_filename(menv, NULL, dest, PATH_MAX, "%s/%s", leader, elem->name);
			if (stat(dest, &sbuf) != 0) {
#ifdef SUPPORT_FILELIST
				if (elem->uid != -1) {
					sbuf.st_uid = elem->uid;
					sbuf.st_gid = elem->gid;
					sbuf.st_mode = elem->mode;
				} else {
#endif
				if (stat(elem->source, &sbuf)) {
					add_perr(perrs, PERROR, "[CopyDir] Invalid source file: %s", elem->source);
					return 1;
				}
#ifdef SUPPORT_FILELIST
				}
#endif
				if (!mkdir(dest, sbuf.st_mode & 0777)) {

					chown(src, sbuf.st_uid, sbuf.st_gid);
				} else {
					add_perr(perrs, PERROR, "Error creating dir %s: %d", dest, errno);
					continue;
				}
			}


			if ((newleader = malloc(strlen(leader) + strlen(elem->name) + 2)) == NULL)
				continue;

			sprintf(newleader, "%s/%s", leader, elem->name);
			copy_dir(menv, elem->subdir, newleader, 1, perrs);
			free(newleader);
			break;

		case MPSS_FILE:
			copy_file(menv, elem, leader, perrs);
			break;

		case MPSS_LINK:
			copy_slink(menv, elem, leader, perrs);
			break;

		case MPSS_DNODE:
			copy_node(menv, elem, leader, perrs);
			break;

		case MPSS_PIPE:
			copy_pipe(menv, elem, leader, perrs);
			break;

		case MPSS_SOCK:
			copy_sock(menv, elem, leader, perrs);
			break;
		}

		elem = elem->next;
	}

	return 0;
}

